import json
from typing import Any, Dict, Optional, Tuple, Type

from sageai.config import LogLevel, get_config, get_function_map, set_config
from sageai.services.openai_service import OpenAIService
from sageai.types.abstract_vectordb import AbstractVectorDB
from sageai.utils.inspection_utilities import get_input_parameter_type
from sageai.utils.openai_utilities import get_latest_user_message

__all__ = ["SageAI"]


class SageAI:
    def __init__(
        self,
        *,
        openai_key: str,
        functions_directory: Optional[str] = None,
        vectordb: Optional[Type[AbstractVectorDB]] = None,
        log_level: Optional[LogLevel] = None,
    ):
        if openai_key is None:
            raise Exception("No OpenAI key provided.")

        config_args = {"openai_key": openai_key}

        if functions_directory is not None:
            config_args["functions_directory"] = functions_directory
        if vectordb is not None:
            config_args["vectordb"] = vectordb
        if log_level is not None:
            config_args["log_level"] = LogLevel(log_level)

        set_config(**config_args)
        self.config = get_config()

        self.openai = OpenAIService()
        self.vectordb = self.config.vectordb()

    def index(self):
        self.vectordb.index()

    def chat(self, *args, **kwargs) -> Dict[str, Any]:
        merged = {i: v for i, v in enumerate(args)}
        merged.update(kwargs)

        top_n = merged.pop("top_n") if "top_n" in merged else None

        if merged.get("model") is None:
            raise Exception("No model provided.")

        if merged.get("messages") is None:
            raise Exception("No messages provided.")

        if top_n is None:
            raise Exception("No top_n provided.")

        latest_user_message = get_latest_user_message(merged.get("messages"))
        if latest_user_message is None:
            raise Exception("No user message found.")

        top_functions = self.get_top_n_functions(
            query=latest_user_message["content"], top_n=top_n
        )

        function_name, function_args = self.call_openai(merged, top_functions)
        function_response = self.run_function(name=function_name, args=function_args)

        base_return = dict(name=function_name, args=function_args)

        if "error" in function_response:
            base_return["error"] = function_response["error"]
        else:
            base_return["result"] = function_response

        return base_return

    def get_top_n_functions(self, *, query: str, top_n: int):
        return self.vectordb.search_impl(query=query, top_n=top_n)

    def call_openai(
        self, openai_args: Dict[str, Any], top_functions: list[Dict[str, Any]]
    ) -> Tuple[str, Dict[str, Any]]:
        openai_result = self.openai.chat(**openai_args, functions=top_functions)

        if not openai_result.function_call:
            raise Exception("No function call found in OpenAI response.")

        function_name = openai_result.function_call.name
        function_args = json.loads(openai_result.function_call.arguments)
        return function_name, function_args

    @staticmethod
    def run_function(*, name: str, args: Dict[str, Any]) -> Dict[str, Any]:
        try:
            function_map = get_function_map()
            func = function_map[name]
            function_input_type = get_input_parameter_type(func.function)
            func_args = function_input_type(**args)
            func_result = func(func_args)
            return func_result.dict()
        except Exception as e:
            return dict(error=str(e))



import argparse
import os

import pytest


def main():
    parser = argparse.ArgumentParser(description="Run tests for functions.")
    parser.add_argument("--directory", type=str, help="Name of function to test")
    parser.add_argument(
        "--apikey", type=str, help="OpenAI API key for integration tests"
    )
    parser.add_argument("--unit", action="store_true", help="Run only the unit tests")
    parser.add_argument(
        "--integration", action="store_true", help="Run only the integration tests"
    )
    args = parser.parse_args()

    if args.directory:
        os.environ["TEST_DIRECTORY"] = args.directory
    if args.apikey:
        os.environ["OPENAI_KEY"] = args.apikey

    current_directory = os.path.dirname(os.path.abspath(__file__))

    if not args.integration or args.unit:
        unit_test_path = os.path.join(current_directory, "unit.py")
        pytest.main(["-x", "-s", unit_test_path, "--disable-warnings"])
    if not args.unit or args.integration:
        integration_test_path = os.path.join(current_directory, "integration.py")
        pytest.main(["-x", "-s", integration_test_path, "--disable-warnings"])



from typing import Optional, Type

from pydantic import BaseModel, Field, ValidationError

from sageai.services.defaultvectordb_service import DefaultVectorDBService
from sageai.types.abstract_vectordb import AbstractVectorDB
from sageai.types.log_level import LogLevel
from sageai.utils.file_utilities import generate_functions_map
from sageai.utils.format_config_args import format_config_args


class Config(BaseModel):
    openai_key: str
    functions_directory: Optional[str] = Field(
        "functions", description="The directory of functions."
    )
    vectordb: Optional[Type[AbstractVectorDB]] = Field(
        DefaultVectorDBService, description="VectorDB class reference."
    )
    log_level: Optional[LogLevel] = Field(
        LogLevel.ERROR, description="The desired log level for output."
    )

    class Config:
        arbitrary_types_allowed = True


_config = Config(openai_key="")
function_map = {}


def set_config(**kwargs):
    """Set the configuration parameters."""
    global _config
    global function_map
    try:
        kwargs = format_config_args(**kwargs)
        _config = Config(**kwargs)
        function_map = generate_functions_map()
        return _config
    except ValidationError as e:
        raise ValidationError(f"Invalid configuration: {e}")


def get_config():
    """Retrieve the configuration."""
    return _config


def get_function_map():
    return function_map



import inspect
import json
import os


def get_user_package_path():
    caller_frame = inspect.stack()[4]
    caller_module = inspect.getmodule(caller_frame[0])

    if caller_module is None or not hasattr(caller_module, "__file__"):
        raise ValueError("Cannot determine the path of the caller's module")

    module_path = os.path.abspath(caller_module.__file__)
    return os.path.dirname(module_path)


def inject_env_vars(**kwargs):
    current_directory = os.path.dirname(os.path.abspath(__file__))
    env_file_path = os.path.join(current_directory, "../env.json")

    if not os.path.exists(env_file_path):
        return kwargs

    with open(env_file_path) as f:
        env_vars = json.load(f)

    if env_vars["OPENAI_KEY"]:
        kwargs["openai_key"] = env_vars["OPENAI_KEY"]

    return kwargs


def format_config_args(**kwargs):
    kwargs = inject_env_vars(**kwargs)

    if "functions_directory" not in kwargs:
        kwargs["functions_directory"] = "functions"
    if not os.path.isabs(kwargs["functions_directory"]) and not kwargs[
        "functions_directory"
    ].startswith("./"):
        base_path = get_user_package_path()
        kwargs["functions_directory"] = os.path.join(
            base_path, kwargs["functions_directory"]
        )
    elif kwargs["functions_directory"].startswith("./"):
        base_path = get_user_package_path()
        # Strip './' from the beginning and then join with the base path
        relative_path = kwargs["functions_directory"][2:]
        kwargs["functions_directory"] = os.path.join(base_path, relative_path)
    return kwargs



from typing import Any, Dict, List, Literal, Optional, Union

from openai import OpenAI
from openai._types import NOT_GIVEN


class OpenAIService:
    def __init__(self):
        from sageai.config import get_config

        config = get_config()

        self.client = OpenAI(api_key=config.openai_key)

    def create_embeddings(
        self,
        *,
        input: Union[str, List[str], List[int], List[List[int]]],
        model: Union[str, Literal["text-embedding-ada-002"]],
        encoding_format: Optional[Literal["float", "base64"]] = NOT_GIVEN,
        user: Optional[str] = NOT_GIVEN,
        **kwargs,
    ) -> List[float]:
        response = self.client.embeddings.create(
            input=input,
            model=model,
            encoding_format=encoding_format,
            user=user,
            **kwargs,
        )
        embeddings = response.data[0].embedding
        return embeddings

    def chat(
        self,
        *,
        messages: List[Dict[str, str]],
        model: str,
        function_call: Optional[Dict[str, Any]] = NOT_GIVEN,
        functions: Dict[str, Any] = NOT_GIVEN,
        max_tokens: Optional[int] = NOT_GIVEN,
        response_format: Optional[Literal["string", "json"]] = NOT_GIVEN,
        temperature: Optional[float] = NOT_GIVEN,
        **kwargs,
    ) -> Dict[str, Any]:
        response = self.client.chat.completions.create(
            messages=messages,
            model=model,
            function_call=function_call,
            functions=functions,
            max_tokens=max_tokens,
            response_format=response_format,
            temperature=temperature,
            **kwargs,
        )
        response_message = response.choices[0].message
        return response_message

    @staticmethod
    def get_embeddings_size():
        return 1536



from abc import ABC, abstractmethod
from typing import Any, Dict, List


class AbstractVectorDB(ABC):
    def __init__(self):
        from sageai.config import get_function_map

        self.function_map = get_function_map()

    @abstractmethod
    def index(self) -> None:
        """Indexes the vector db based on the functions directory."""
        pass

    @abstractmethod
    def search(self, *, query: str, top_n: int) -> List[str]:
        """Actual search logic, which should be implemented in derived classes.
        It should return a list of function names
        """
        pass

    def format_search_result(
        self, *, function_names: List[str]
    ) -> List[Dict[str, Any]]:
        potential_functions = [
            self.function_map[func_name].parameters for func_name in function_names
        ]
        return potential_functions

    def search_impl(self, *, query: str, top_n: int) -> List[Dict[str, Any]]:
        """Search vector db based on a query and return top n function names."""
        results = self.search(query=query, top_n=top_n)
        return self.format_search_result(function_names=results)



import logging
from sageai.types.log_level import LogLevel


def get_logger(logger_name: str, log_level: LogLevel):
    logger = logging.getLogger(logger_name)

    log_level_value = log_level.value
    logger.setLevel(log_level_value)

    formatter = logging.Formatter("%(asctime)s | [%(levelname)s] %(message)s")

    if not logger.handlers:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(log_level_value)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
    else:
        logger.handlers[0].setLevel(log_level_value)  # assumes only one handler

    return logger



import json
import os
from time import sleep
from typing import cast, get_type_hints
import pydantic

import pytest

from sageai.sageai import SageAI
from sageai.types.function import Function
from sageai.types.log_level import LogLevel
from sageai.utils.file_utilities import get_functions_directories, load_module_from_file
from sageai.utils.logger import get_logger

logger = get_logger("IntegrationTest", LogLevel.INFO)


@pytest.mark.parametrize(
    "dirpath",
    get_functions_directories(
        logger=logger, functions_directory_path=os.getenv("TEST_DIRECTORY")
    ),
)
def test_integration(dirpath):
    sleep(2)
    test_file = os.path.join(dirpath, "test.json")
    function_file = os.path.join(dirpath, "function.py")
    function_module = load_module_from_file("function", function_file)
    function_to_test = cast(Function, getattr(function_module, "function"))

    directory_full_path = os.path.abspath(os.getcwd())
    functions_directory = os.path.join(directory_full_path, dirpath)

    openai_key = os.environ["OPENAI_KEY"]
    sageai = SageAI(openai_key=openai_key, functions_directory=functions_directory)
    sageai.index()

    try:
        output_data_model_class = get_type_hints(function_to_test.function)["return"]

        with open(test_file) as f:
            test_cases = json.load(f)
            for i, test_case in enumerate(test_cases):
                logger.info(
                    f"{function_module.function.name} - Running test case {i+1}/{len(test_cases)}: {test_case['message']}",
                )
                try:
                    vector_db_result = sageai.vectordb.search(
                        query=test_case["message"], top_n=5
                    )
                    result = sageai.chat(
                        messages=[dict(role="user", content=test_case["message"])],
                        model="gpt-3.5-turbo-0613",
                        top_n=5,
                    )
                    if result is None:
                        logger.error(
                            "No result returned from Sennin for function: {}".format(
                                function_module.function.name
                            )
                        )
                        continue

                    test_output = None
                    try:
                        test_output = output_data_model_class(
                            **test_case["output"]
                        ).dict()
                    except pydantic.ValidationError as ve:
                        expected_fields = ", ".join(
                            get_type_hints(output_data_model_class).keys()
                        )
                        actual_fields = test_case["output"]
                        logger.error(
                            f"Validation error for function output {function_module.function.name}:\nExpected fields: {expected_fields}\nGot: {actual_fields}.\n{str(ve)}"
                        )
                        continue

                    if function_module.function.name not in vector_db_result:
                        logger.error(
                            f"VectorDB did not return expected function for {function_module.function.name}. Expected: {function_module.function.name} in {vector_db_result}."
                        )
                        continue
                    if result["name"] != function_module.function.name:
                        logger.error(
                            f"Wrong function chosen by model for {function_module.function.name}. Expected: {function_module.function.name}, Got: {result['name']}"
                        )
                        continue
                    if result["result"] != test_output:
                        logger.error(
                            f"Output mismatch for {function_module.function.name}. Expected: {test_output}, Got: {result['result']}"
                        )
                        continue

                except Exception as e:
                    logger.error(
                        f"Function {function_module.function.name} execution failed: {str(e)}"
                    )

    except Exception as e:
        logger.error(f"Entire test failed: {str(e)} for directory {dirpath}")



import json
import os
from typing import cast, get_type_hints
import pydantic

import pytest

from sageai.types.function import Function
from sageai.types.log_level import LogLevel
from sageai.utils.file_utilities import get_functions_directories, load_module_from_file
from sageai.utils.logger import get_logger

logger = get_logger("UnitTest", LogLevel.INFO)


def assert_function_module_correct(folder_name, function_module):
    if not hasattr(function_module, "function"):
        msg = f"{folder_name}/function.py does not export a variable called 'function'"
        raise ValueError(msg)


def assert_test_cases_correct(test_cases, folder_name):
    for idx, test_case in enumerate(test_cases):
        if not isinstance(test_case, dict):
            msg = f"Test case #{idx + 1} in {folder_name}/test.json is not an object"
            raise ValueError(msg)

        missing_properties = [
            prop for prop in ["input", "output", "message"] if prop not in test_case
        ]
        if missing_properties:
            msg = f"Test case #{idx + 1} in {folder_name}/tests.json missing properties: {', '.join(missing_properties)}"
            raise ValueError(msg)

        if not isinstance(test_case.get("message", ""), str):
            msg = f"Test case #{idx + 1} in {folder_name}/tests.json has a 'message' property which is not of type string"
            raise ValueError(msg)


@pytest.mark.parametrize(
    "dirpath",
    get_functions_directories(
        logger, functions_directory_path=os.getenv("TEST_DIRECTORY")
    ),
)
def test_unit(dirpath):
    folder_name = os.path.basename(dirpath)
    function_file = os.path.join(dirpath, "function.py")
    test_file = os.path.join(dirpath, "test.json")

    try:
        # Check necessary files
        if not os.path.exists(function_file):
            msg = f"Missing function.py in {folder_name}"
            raise ValueError(msg)

        if not os.path.exists(test_file):
            msg = f"Missing test.json in {folder_name}"
            raise ValueError(msg)

        function_module = load_module_from_file("function", function_file)
        function_to_test = cast(Function, getattr(function_module, "function"))
        assert_function_module_correct(folder_name, function_module)

        input_model_class = function_to_test.input_type
        output_data_model_class = get_type_hints(function_to_test.function)["return"]

        with open(test_file) as f:
            test_cases = json.load(f)
            assert_test_cases_correct(test_cases, folder_name)

            for test_case in test_cases:
                logger.info(f"Running message {test_case['message']}")
                func_output = function_to_test(input_model_class(**test_case["input"]))

                test_output = None
                try:
                    test_output = output_data_model_class(**test_case["output"]).dict()
                except pydantic.ValidationError as ve:
                    expected_fields = ", ".join(
                        get_type_hints(output_data_model_class).keys()
                    )
                    actual_fields = test_case["output"]
                    logger.error(
                        f"Validation error for function output {function_module.function.name}:\nExpected fields: {expected_fields}\nGot: {actual_fields}.\n{str(ve)}"
                    )
                    continue

                if func_output != test_output:
                    logger.error(
                        f"Function {folder_name} output mismatch for message '{test_case['message']}':"
                    )
                    logger.error(f"Expected: {test_output}")
                    logger.error(f"Got: {func_output}")
                    raise AssertionError("Expected output does not match actual output")

    except Exception as e:
        logger.error(f"Function {folder_name} failed. Error: {str(e)}")



from typing import Dict, List


def get_latest_user_message(messages: List[Dict]) -> Dict | None:
    for message in reversed(messages):
        if message.get("role") == "user":
            return message
    return None



import re
from typing import List

from lasso_sennin.utils.address_mapping_utilities import test_address_mappings


def replace_with_index(string: str, pattern: str, replacement_list: List[str]):
    for match in re.finditer(pattern, string):
        index = int(match.group(1)) - 1
        string = string.replace(match.group(0), replacement_list[index])
    return string


def replace_without_index(string: str, pattern: str, replacement: str):
    return string.replace(pattern, replacement)


def replace_addresses_with_test_addresses(string: str):
    for key, value in test_address_mappings.items():
        pattern_base = re.escape(
            key[:-1],
        )
        string = replace_with_index(
            string,
            pattern_base + r"_(\d+)>",
            value,
        )
        string = replace_without_index(
            string,
            pattern_base + r">",
            value[0],
        )
    return string
